module fft_data_mod

  use down_process_types
  use data_types
  use timers_mod

  implicit none

  integer, parameter ::  FFT_FORW=1
  integer, parameter ::  FFT_BACK=-1

  type fft_type 
     integer :: ierr
     real :: scale
     complex, pointer, dimension (:) :: fft_coeff
  end type fft_type

  integer, intrinsic :: omp_get_thread_num


contains

  function  init_fft_1d(n1,fast_init) result (fft_id)
    type (fft_type) fft_id
    integer n1,stat
    logical, optional :: fast_init

    allocate (fft_id%fft_coeff(n1+15),stat=stat)  
    fft_id%ierr =stat
    if(fft_id%ierr .ne. 0) call seperr("Problems with allocation of fft_coeff (1-D)")
    call cfft1di(n1,fft_id%fft_coeff)
    fft_id%scale=sqrt(1./(n1))

  end function init_fft_1d

  !----------------------------------------------------------------------
  function  init_fft_2d(n1,n2,fast_init) result (fft_id)
    type (fft_type) fft_id
    integer n1,n2,stat
    logical, optional :: fast_init

    allocate (fft_id%fft_coeff(n1+n2+30),stat=stat)  
    fft_id%ierr =stat
    if(fft_id%ierr .ne. 0) call seperr("Problems with allocation of fft_coeff (2-D)")

    call cfft2di(n1,n2,fft_id%fft_coeff)
    fft_id%scale=sqrt(1./(n1*n2))

    !write(0,*)'shape(fft_coeff)=',shape(fft_id%fft_coeff)
    !write(0,*)'scale=',fft_id%scale

  end function init_fft_2d

  !----------------------------------------------------------------------
  function init_fft_3d(n1,n2,n3,fast_init) result (fft_id)
    type (fft_type) fft_id
    integer n1,n2,n3,stat
    logical, optional :: fast_init
    character (len=128) :: desc_timer

    allocate (fft_id%fft_coeff(n1+n2+n3+45),stat=stat)  
    fft_id%ierr =stat
    if(fft_id%ierr .ne. 0) call seperr("Problems with allocation of fft_coeff (3-D)")

    desc_timer='Init FFTW 3-D (SGI)'
    call SEP_timer_init(20,desc_timer,0.,0)

    call SEP_timer_start(20)
    call cfft3di(n1,n2,n3,fft_id%fft_coeff)
    fft_id%scale=sqrt(1./(n1*n2*n3))
    call SEP_timer_stop(20)

    desc_timer='Computing FFTW 3-D (SGI)'
    call SEP_timer_init(21,desc_timer,0.,0)

  end function init_fft_3d

  !----------------------------------------------------------------------
  subroutine fft_1d_data(fft_id,sign,n1,data_slice)
    type (fft_type) fft_id
    integer :: n1
    integer :: sign
    complex, dimension (:) :: data_slice

    call cscal1d(n1,fft_id%scale,data_slice,1)
    call cfft1d(sign,n1,data_slice,1,fft_id%fft_coeff)

    return	
  end subroutine fft_1d_data


  !----------------------------------------------------------------------
  subroutine fft_2d_data(fft_id,sign,n1,n2,data_slice)
    type (fft_type) fft_id
    integer :: n1,n2
    integer :: sign
    complex, dimension (:,:) :: data_slice

    call cscal2d(n1,n2,fft_id%scale,data_slice,n1)
    call cfft2d(sign,n1,n2,data_slice,n1,fft_id%fft_coeff)


    return	
  end subroutine fft_2d_data


  !----------------------------------------------------------------------
  subroutine fft_3d_data(fft_id,sign,n1,n2,n3,data_slice)
    type (fft_type) fft_id
    integer :: n1,n2,n3,i1,i2,i3
    integer :: sign
    complex,  dimension (:,:,:) :: data_slice

    call cscal3d(n1,n2,n3,fft_id%scale,data_slice,n1,n2)
    if(omp_get_thread_num() == 0) then
       call SEP_timer_start(21)
    end if
    call cfft3d(sign,n1,n2,n3,data_slice,n1,n2,fft_id%fft_coeff)
    if(omp_get_thread_num() == 0) then
       call SEP_timer_stop(21)
    end if


    return	
  end subroutine fft_3d_data

  !----------------------------------------------------------------------
  subroutine fft_3d_two_pass_data(fft_1d_id,fft_2d_id,sign,n1,n2,n3,data_slice)
    type (fft_type) fft_1d_id,fft_2d_id
    integer :: n1,n2,n3,i1,i2,i3
    integer :: sign,inc
    real    :: scale
    complex,  dimension (:,:,:) :: data_slice

    ! write(0,*)'fft_3d sign=',sign
    ! write(0,*)'fft_3d n1=',n1
    ! write(0,*)'fft_3d n2=',n2
    ! write(0,*)'fft_3d n3=',n3
    ! write(0,*)'fft_3d shape(data_slice)=',shape(data_slice)

    scale=fft_2d_id%scale*fft_1d_id%scale

    !write(0,*)'n1=',n1
    !write(0,*)'n2=',n2
    !write(0,*)'n3=',n3
    !write(0,*)'scale=',scale
    !write(0,*)'shape(fft_1d_id-coeff)=',shape(fft_1d_id%fft_coeff)
    !write(0,*)'shape(fft_2d_id-coeff)=',shape(fft_2d_id%fft_coeff)

    do i3=1,n3
       call cscal2d(n1,n2,scale,data_slice(:,:,i3),n1)
       call cfft2d(sign,n1,n2,data_slice(:,:,i3),n1,fft_2d_id%fft_coeff)
    end do

    if(n3 > 1) then
       inc=n1*n2
       do i2=1,n2
          do i1=1,n1
             call cfft1d(sign,n3,data_slice(i1,i2,1),inc,fft_1d_id%fft_coeff)
          end do
       end do
    end if

    return	
  end subroutine fft_3d_two_pass_data

  !----------------------------------------------------------------------
  subroutine fft_4d_two_pass_data(fft_1d_id,fft_3d_id,sign, &
  n1,n2,n3,n4,data_slice)
    type (fft_type) fft_1d_id,fft_3d_id
    integer :: n1,n2,n3,n4,i1,i2,i3,i4
    integer :: sign,inc
    real    :: scale
    complex,  dimension (:,:,:,:) :: data_slice

    ! write(0,*)'fft_4d sign=',sign
    ! write(0,*)'fft_4d n1=',n1
    ! write(0,*)'fft_4d n2=',n2
    ! write(0,*)'fft_4d n3=',n3
    ! write(0,*)'fft_4d n4=',n4
    ! write(0,*)'fft_4d shape(data_slice)=',shape(data_slice)

    scale=fft_1d_id%scale*fft_3d_id%scale

    !write(0,*)'scale=',scale
    !write(0,*)'shape(fft_1d_id-coeff)=',shape(fft_1d_id%fft_coeff)
    !write(0,*)'shape(fft_3d_id-coeff)=',shape(fft_3d_id%fft_coeff)

    do i4=1,n4
       call cscal3d(n1,n2,n3,scale,data_slice(:,:,:,i4),n1,n2)
       call cfft3d(sign,n1,n2,n3,data_slice(:,:,:,i4),n1,n2,fft_3d_id%fft_coeff)
    end do

    if(n4 > 1) then
       inc=n1*n2*n3
       do i3=1,n3
          do i2=1,n2
             do i1=1,n1
                call cfft1d(sign,n4,data_slice(i1,i2,i3,1),inc,fft_1d_id%fft_coeff)
             end do
          end do
       end do
    end if

    return	
  end subroutine fft_4d_two_pass_data


  !----------------------------------------------------------------------
  subroutine clean_fft(fft_id) 
    type (fft_type) fft_id
    integer ierr

    !if(associated(fft_id%fft_coeff)) nullify (fft_id%fft_coeff)  
    if(associated(fft_id%fft_coeff)) then
       deallocate (fft_id%fft_coeff,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocation of fft_coeff')
    end if

    return	
  end subroutine clean_fft

end module fft_data_mod
