module NAM_mod
  
! Type definitions for this module
use NAM_types

! Types definition for Gendown package
use gendown3d_types
use down_process_types
use data_types
use image_types
use slow_types

! FFT's module
use fft_data_mod

! Boundary condition module
use bound_mod

implicit none

type (NAM_param_type), private,save :: NAM_par_loc
type (fft_type), private,save :: fft_1d_id,fft_3d_id
logical, private,save :: apply_any_bound
real, private,save :: slow_const,slow_salt,epsi,min_prop_cos_sq
real, private,save :: d_k_norm,min_lap_sq,max_disp_rel
real, allocatable, dimension (:), private,save :: magn,phase
real, allocatable, dimension (:), private,save :: k_mx,k_hx,k_my,kk_hy
real, allocatable, dimension (:,:), private,save :: k_r_x_sq,k_s_x_sq
!real, allocatable, dimension (:,:), private :: k_r_y_sq,k_s_y_sq

contains

!-----------------------------------------------------------------------------
! copies parameters into a structure private to the module 
subroutine init_NAM_par(NAM_par)
  type (NAM_param_type) NAM_par
  
  NAM_par_loc=NAM_par

end subroutine init_NAM_par

!-----------------------------------------------------------------------------
! Initiliazes parameters and allocate and initialize arrays
function NAM_init(gen_par,data_par,slow_par)  result(info)

  type (gendown3d_info_type) info
  type (gendown3d_param_type) gen_par
  type (data_param_type) data_par
  type (slow_param_type) slow_par

  integer i_k_mx,i_k_my,i_k_hx,i_k_hy
  integer ierr,i_k_norm
  real kmx,kmy,khx,khy,pi2,k_norm_sq,k_norm
  real n_cycles_ksq,k_start,k_cutoff,k_decay,sigma_magn_sq,edge_gauss,gauss_magn
  real delta_k

!info=info_default already initialized in declaration of type

  ! constant slowness
  slow_const=gen_par%slow_const

  ! set pi and epsi
  pi2=2.*acos(-1.)
  epsi=epsilon(1.)

  ! set maximum propagation angle 
  min_prop_cos_sq=cos(gen_par%max_propag_angle*pi2/360.)**2

  apply_any_bound=gen_par%apply_bound

  ! check paprameter
  if((NAM_par_loc%mig_type(1:3) == 'cam') .and. (data_par%n_hy_pad > 1)) then
    call seperr('If mig_type equal to CAM, n_hy_pad must be equal to 1') 
  end if

  ! set slow_salt to positive if defined, otherwise to negative
  if (slow_par%slowness) then
     if(NAM_par_loc%slow_salt > 0.) then
        slow_salt=1.01*(NAM_par_loc%slow_salt)
        if(NAM_par_loc%n_slow_split < 2) call seperr('n_slow_split <2')
     else 
        slow_salt=-1.
     end if
  else
     if(NAM_par_loc%vel_salt > 0.) then
        slow_salt=1.01*(1./NAM_par_loc%vel_salt)
        if(NAM_par_loc%n_slow_split < 2) call seperr('n_slow_split <2')
     else 
        slow_salt=-1.
     end if
  end if


  ! initialize FFTs
  if(apply_any_bound) then
    call init_fft_1d(fft_1d_id,data_par%n_hy_pad,fast_init=.false.)
!     fft_1d_id=init_fft_1d(data_par%n_hy_pad,fast_init=.false.)
    info%status=fft_1d_id%ierr
    call init_fft_3d(fft_3d_id,data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,fast_init=.false.)
!    fft_3d_id=init_fft_3d(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,fast_init=.false.)
    info%status=fft_3d_id%ierr

!write(0,*)'NAM_mod scale 1d=',fft_1d_id%scale
!write(0,*)'NAM_mod scale 1d=',fft_1d_id%ndim
!write(0,*)'NAM_mod scale 1d=',fft_1d_id%ierr
!    write(0,*)'scale=',fft_3d_id%scale
   end if

  ! check that there are no cross-line offsets
  if((data_par%n_hy_pad /= 1 ).and.(NAM_par_loc%mig_type(1:3) /= 'fps')) then
     write(0,*) "NAM: Trying to use NAM migration with n_hy > 1  (",&
     data_par%n_hy_pad,")"
  end if

  ! allocate tables for square-root operators 
  allocate (phase(NAM_par_loc%n_disp_rel),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of phase ")
  allocate (magn(NAM_par_loc%n_disp_rel),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of magn ")

  ! allocate tables for wavenumbers
  allocate (k_mx(data_par%n_mx_pad),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
  allocate (k_my(data_par%n_my_pad),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of k_my ")
  allocate (k_hx(data_par%n_hx_pad),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")
  allocate (kk_hy(data_par%n_hy_pad),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of kk_hy ")

  allocate (k_r_x_sq(data_par%n_mx_pad,data_par%n_hx_pad), &
  stat=ierr)
  if(ierr .ne. 0) call seperr("Problems with allocation of k_r_x_sq ")
  allocate (k_s_x_sq(data_par%n_mx_pad,data_par%n_hx_pad), &
  stat=ierr)
  if(ierr .ne. 0) call seperr("Problems with allocation of k_s_x_sq ")

!  allocate (k_r_y_sq(data_par%n_my_pad,data_par%n_hy_pad), &
!  stat=ierr)
!  if(ierr .ne. 0) call seperr("Problems with allocation of k_r_y_sq ")
!  allocate (k_s_y_sq(data_par%n_my_pad,data_par%n_hy_pad), &
!  stat=ierr)
!  if(ierr .ne. 0) call seperr("Problems with allocation of k_s_y_sq ")

!  write(0,*) 'wavenumber dimensions',data_par%n_mx_pad,data_par%n_my_pad,&
!  Data_par%n_hx_pad,data_par%n_hy_pad

!  write(0,*)'d_kmx=',data_par%d_kmx
!  write(0,*)'d_kmy=',data_par%d_kmy
!  write(0,*)'kmx_0=',data_par%kmx_0
!  write(0,*)'kmy_0=',data_par%kmy_0
!  write(0,*)'d_khx=',data_par%d_khx
!  write(0,*)'d_khy=',data_par%d_khy
!  write(0,*)'khx_0=',data_par%khx_0
!  write(0,*)'khy_0=',data_par%khy_0

  ! init k_mx table
  do i_k_mx=1,data_par%n_mx_pad
     kmx  = i_k_mx - 1.
     if(kmx <= float(data_par%n_mx_pad/2)) then
        k_mx(i_k_mx) =  data_par%kmx_0+data_par%d_kmx* kmx
     else
        k_mx(i_k_mx) = -data_par%kmx_0+data_par%d_kmx*(kmx - data_par%n_mx_pad)
     end if
  end do

  ! init k_my table
  do i_k_my=1,data_par%n_my_pad
     kmy = i_k_my - 1.
     if(kmy <= float(data_par%n_my_pad/2)) then
        k_my(i_k_my) =  data_par%kmy_0+data_par%d_kmy* kmy
     else
        k_my(i_k_my) = -data_par%kmy_0+data_par%d_kmy*(kmy - data_par%n_my_pad)
     end if
  end do

  ! init k_hx table
  if(gen_par%single_hx_dip) then
     do i_k_hx=1,data_par%n_hx_pad
        khx = i_k_hx - 1.
        k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
     end do
  else
     do i_k_hx=1,data_par%n_hx_pad
        khx = i_k_hx - 1.
        if(khx <= float(data_par%n_hx_pad/2)) then
           k_hx(i_k_hx) =  data_par%khx_0+data_par%d_khx* khx
        else
           k_hx(i_k_hx) = -data_par%khx_0+data_par%d_khx*(khx - data_par%n_hx_pad)
        end if
     end do
  end if

  ! init kk_hy table
  do i_k_hy=1,data_par%n_hy_pad
     khy = i_k_hy - 1.
     if(khy <= float(data_par%n_hy_pad/2)) then
        kk_hy(i_k_hy) =  khy
     else
        kk_hy(i_k_hy) = -data_par%n_hy_pad + khy 
     end if
  end do


  ! init source and geophone components of wavenumbers
  do i_k_hx=1,data_par%n_hx_pad
     do i_k_mx=1,data_par%n_mx_pad
        k_s_x_sq(i_k_mx,i_k_hx)= &
        .25*((k_mx(i_k_mx) - k_hx(i_k_hx))**2)
        k_r_x_sq(i_k_mx,i_k_hx)= &
        .25*((k_mx(i_k_mx) + k_hx(i_k_hx))**2)
     end do
  end do

!  do i_k_hy=1,data_par%n_hy_pad
!     do i_k_my=1,data_par%n_my_pad
!        k_s_y_sq(i_k_my,i_k_hy)= &
!        .25*((k_my(i_k_my) - k_hy(i_k_hy))**2)
!        k_r_y_sq(i_k_my,i_k_hy)= &
!        .25*((k_my(i_k_my) + k_hy(i_k_hy))**2)
!     end do
!  end do

  ! now initialize square-root operator tables
  n_cycles_ksq=5.
  k_cutoff =0.
  delta_k=sqrt(min_prop_cos_sq)
  k_start=delta_k
  k_decay=-1.*delta_k
  if(k_decay**2 > n_cycles_ksq) call seperr('k_decay^2 > n_cycles_ksq')
  edge_gauss = exp(k_decay)
  sigma_magn_sq=-(k_decay-k_start)**2/(2.*log(edge_gauss))

  d_k_norm=n_cycles_ksq/(NAM_par_loc%n_disp_rel-1)
  do i_k_norm=1,NAM_par_loc%n_disp_rel
     k_norm_sq=1.-(i_k_norm-1)*d_k_norm
     if( k_norm_sq > 0.) then
        k_norm=sqrt(k_norm_sq)
     else
        k_norm=-sqrt(-k_norm_sq)
     end if

     phase(i_k_norm)=k_norm

     if(k_norm > k_start) then
        magn(i_k_norm)=0.
     else
        magn(i_k_norm)=(-((k_norm-k_start)**2)/(2.*sigma_magn_sq))
!        magn(i_k_norm)=-10000000000000
     end if
  end do

  max_disp_rel=float(NAM_par_loc%n_disp_rel)


    call snap('Phase.h',NAM_par_loc%n_disp_rel,1,phase)
    call snap('Magn.h',NAM_par_loc%n_disp_rel,1,magn)

end function NAM_init

!-----------------------------------------------------------------------------
! downward continue wavefield by one depth step
recursive function NAM_down_oper(adj,add, &
     n_w_data,w_0_data,d_w_data, &
     z_down,d_z_down,data_par, &
     data_slice,ddata_slice,slow_slice) &
     result(stat)


logical, intent(in) :: add,adj
integer, intent(in) :: n_w_data
real, intent(in) :: w_0_data,d_w_data,z_down,d_z_down
type (data_param_type), intent(in) :: data_par
real, dimension (:,:,:,:,:), intent(inout) :: slow_slice
complex, dimension (:,:,:,:,:), intent(inout) :: data_slice,ddata_slice
optional data_slice,ddata_slice,slow_slice

integer :: stat

logical, automatic :: found_salt
logical, automatic :: prop_data,prop_ddata,var_slow,prop_salt
integer, automatic :: i_k_mx,i_k_my,i_k_hx,i_k_hy,n_mx,n_my,n_hx,n_hy,i_mx,i_my,i_hx,i_hy
integer, automatic :: ierr,i_ksq_norm,i_ksq_norm_s,i_ksq_norm_r,i_w,alg_dist
integer, automatic :: i_slow,n_slow_split_eff,n_slow_split_no_salt,map_slow
real, automatic :: k0_sq,rphase,rphase_s,rphase_r,rmagn,dphase,w_data,k_sq_comaz
real, automatic :: min_slow,max_slow,delta_slow,slow_0,inv_delta_slow,range_slow,mean_slow
real, automatic :: sigma,rat,dscat,dz_w,pi2,norm_ksq,k0,k0_dz,slow
real, automatic :: k_hy,k_hy_min,k_hy_max,k_hy_mean,d_k_hy
real, automatic :: norm_ksq_min_alpha,norm_ksq_max_alpha
complex, automatic :: ddscat

!complex, allocatable,  dimension (:,:,:,:) :: temp_data_1

real, automatic, dimension(NAM_par_loc%n_slow_split)    :: slow_prop
!real,  dimension(data_par%n_mx_pad,data_par%n_hx_pad) :: alpha_s,alpha_r
real,automatic,  dimension(data_par%n_mx_pad,data_par%n_hx_pad) :: alpha_min,alpha_max
complex,automatic, dimension (NAM_par_loc%n_disp_rel) :: ph_shift_tab

integer, external :: omp_get_thread_num

! set constants
n_mx=data_par%n_mx_pad
n_my=data_par%n_my_pad
n_hx=data_par%n_hx_pad
n_hy=data_par%n_hy_pad
pi2=2.*acos(-1.)

! if data_slice is present propagate background wavefield
prop_data= present(data_slice)

! if ddata_slice is present propagate scattered wavefield
prop_ddata= present(ddata_slice)

! if slow_slice is present use variable velocity, otherwise use slow_const
var_slow= present(slow_slice)


! allocate temporary arrays for wavefield copies
if(prop_data) then
!   allocate(temp_data_1(data_par%n_mx_pad,data_par%n_my_pad,data_par%n_hx_pad,&
!        data_par%n_hy_pad),stat=ierr)
!   if(ierr .ne. 0) then
!      write(0,*) "Tried to allocate temp_data_1 with shape",shape(temp_data_1)
!      call seperr("Problems with allocation of temp_data_1")
!   endif
end if

if(prop_ddata) then
  call seperr('NAM does not support scattering yet')
end if


if(var_slow) then
   !{
   ! set  min_slow, max_slow  and n_slow_split_no_salt 
   if(slow_salt > 0.) then 
      ! salt slowness was defined 
      min_slow=minval(slow_slice,mask=(slow_slice > slow_salt))
      found_salt=any(slow_slice <= slow_salt)
      if(found_salt) then 
         ! salt found in this depth slice
         n_slow_split_no_salt=NAM_par_loc%n_slow_split-1
      else
         ! salt NOT found in this depth slice
         n_slow_split_no_salt=NAM_par_loc%n_slow_split
      end if
   else
      ! salt slowness was NOT defined 
      found_salt=.false.
      min_slow=minval(slow_slice)
      n_slow_split_no_salt=NAM_par_loc%n_slow_split
   end if
   max_slow=maxval(slow_slice)
   
   ! set  inv_delta_slow and n_slow_split_no_salt 
   range_slow=max_slow-min_slow
   mean_slow=.5*(max_slow+min_slow)
   delta_slow=range_slow/n_slow_split_no_salt
   if(delta_slow < NAM_par_loc%min_delta_slow*mean_slow) then
      if(range_slow < NAM_par_loc%min_delta_slow*mean_slow) then
         n_slow_split_no_salt= 1 
         delta_slow=range_slow
         inv_delta_slow=0.
      else
         n_slow_split_no_salt= & 
              min(n_slow_split_no_salt, &
              max(1,int(range_slow/(NAM_par_loc%min_delta_slow*mean_slow))))
         delta_slow=range_slow/n_slow_split_no_salt
         inv_delta_slow=1./delta_slow
      end if
   else
      inv_delta_slow=1./delta_slow
   end if

   ! set velocities outside salt
   do i_slow=1,n_slow_split_no_salt
      slow_prop(i_slow)=min_slow+(i_slow-.5)*delta_slow
   end do
   
   ! set  n_slow_split_eff 
   if(found_salt) then
      n_slow_split_eff= n_slow_split_no_salt + 1
      slow_prop(n_slow_split_eff)=slow_salt
   else
      n_slow_split_eff= n_slow_split_no_salt 
   end if
   !}
else
   !{
   if(NAM_par_loc%n_slow_split /= 1) call seperr('n_slow_split /= 1')
   n_slow_split_eff= 1 
   slow_prop(1)=slow_const
   !}
end if


!............................................................................
do i_w=1,n_w_data
   !{ beginning loop over frequencies
   w_data=w_0_data+(i_w-1)*d_w_data

!   do i_k_hy=1,n_hy
!      do i_k_hx=1,n_hx
!         do i_k_my=1,n_my
!            do i_k_mx=1,n_mx
!               temp_data_1(i_k_mx,i_k_my,i_k_hx,i_k_hy) = &
!                    data_slice(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w) 
!            enddo
!         enddo
!      enddo
!   enddo
   
!............................................................................
do i_slow=1,n_slow_split_eff
   !{ beginning loop over slowness for sources
   ! set current slowness
   slow_0=slow_prop(i_slow)

!   write(0,*)'slow_prop=',slow_prop
   write(0,*)'slow_0=',slow_0
   
   ! determine if in salt
   if(found_salt .and. i_slow == n_slow_split_eff) then
      prop_salt=.true.
   else
      prop_salt=.false.
   end if
   
   k0= (w_data*slow_0)
   k0_sq= k0*k0
   k0_dz=k0*d_z_down
   dz_w=d_z_down*w_data
   
   ! compute wavenumber normalization factor to look-up into square-root table
   if(k0_sq > epsi) then
      norm_ksq=1./(d_k_norm*k0_sq)
      if((NAM_par_loc%mig_type(1:3) == 'cam')) then
        norm_ksq_min_alpha=norm_ksq
        norm_ksq_max_alpha=norm_ksq
      else
! BB - changes MAy 200
        norm_ksq_min_alpha= &
               1./(d_k_norm*((w_data*slow_0*NAM_par_loc%min_slow_alpha)**2))
        norm_ksq_max_alpha= &
               1./(d_k_norm*((w_data*slow_0*NAM_par_loc%max_slow_alpha)**2))
!        norm_ksq_min_alpha= &
!               1./(d_k_norm*((w_data*NAM_par_loc%min_slow_alpha)**2))
!        norm_ksq_max_alpha= &
!               1./(d_k_norm*((w_data*NAM_par_loc%max_slow_alpha)**2))
      end if
   else
      write(0,*)'I should not be here k0_sq=',k0_sq
      write(0,*)'k0_sq=',k0_sq
      write(0,*)'slow_0=',slow_0
      write(0,*)'w_data=',w_data
      call seperr('NAM: Zero frequency')
      norm_ksq=2.*NAM_par_loc%n_disp_rel/min_lap_sq
   end if

   if(k0_dz >0) then
      ph_shift_tab(:)=cexp(k0_dz*cmplx(magn(:),phase(:)))
   else
      ph_shift_tab(:)=cexp(k0_dz*cmplx(magn(:),-phase(:)))
   end if
   
   if((var_slow) .and. (adj)) then
      !{
      ! not needed for migration
      call seperr('wrong option  1')
      !}
   end if

   if(NAM_par_loc%mig_type(1:3) .ne. 'fps') then
   !{
   ! compute phase shift at the sources for current slowness using tables
   do i_k_hx=1,n_hx
      do i_k_mx=1,n_mx
         ! in-line source wavenumber
         i_ksq_norm=nint(min(k_s_x_sq(i_k_mx,i_k_hx)*norm_ksq_min_alpha+1,max_disp_rel))
         rphase_s=(phase(i_ksq_norm))
         ! in-line receiver wavenumber
         i_ksq_norm=nint(min(k_r_x_sq(i_k_mx,i_k_hx)*norm_ksq_min_alpha+1,max_disp_rel))
         rphase_r=(phase(i_ksq_norm))
         if((rphase_s > 0.) .and. (rphase_r > 0.)) then
            ! apply common-azimuth equation if propagating waves
            alpha_min(i_k_mx,i_k_hx)=(rphase_r-rphase_s)/(rphase_r+rphase_s)
         else
            ! set k_hy to 0 if evanescent.
            alpha_min(i_k_mx,i_k_hx)=0.
         end if

         ! in-line source wavenumber
         i_ksq_norm=nint(min(k_s_x_sq(i_k_mx,i_k_hx)*norm_ksq_max_alpha+1,max_disp_rel))
         rphase_s=(phase(i_ksq_norm))
         ! in-line receiver wavenumber
         i_ksq_norm=nint(min(k_r_x_sq(i_k_mx,i_k_hx)*norm_ksq_max_alpha+1,max_disp_rel))
         rphase_r=(phase(i_ksq_norm))
         if((rphase_s > 0.) .and. (rphase_r > 0.)) then
            ! apply common-azimuth equation if propagating waves
            alpha_max(i_k_mx,i_k_hx)=(rphase_r-rphase_s)/(rphase_r+rphase_s)
         else
            ! set k_hy to 0 if evanescent.
            alpha_max(i_k_mx,i_k_hx)=0.
         end if


      end do
      if(i_k_hx == data_par%n_hx_pad/2 + 1) then
         alpha_min(:,i_k_hx)=0.
         alpha_max(:,i_k_hx)=0.
      end if
   end do
  !}
  end if

   
   do i_k_hx=1,n_hx
   !{
   do i_k_my=1,n_my
   !{
   do i_k_mx=1,n_mx
   !{

    select case (NAM_par_loc%mig_type(1:3)) 
    case ('cam')
      k_hy_mean=alpha_max(i_k_mx,i_k_hx)*k_my(i_k_my)
      d_k_hy=0.
    case ('nam')
      k_hy_min=(alpha_max(i_k_mx,i_k_hx)*k_my(i_k_my))
      k_hy_max=(alpha_min(i_k_mx,i_k_hx)*k_my(i_k_my))
      if(NAM_par_loc%even_sampling) then
        k_hy_mean=k_hy_min + .5*(k_hy_max - k_hy_min)
        d_k_hy=data_par%d_khy
      else
        if(n_hy > 1) then
          d_k_hy=(k_hy_max-k_hy_min)/(n_hy-1)
        else
          d_k_hy=0.
        end if
        k_hy_mean=k_hy_max - d_k_hy*n_hy/2
      end if
    case ('fps')
      d_k_hy=data_par%d_khy
      k_hy_mean=0.
    case default
      write(0,*)'mig_type=',NAM_par_loc%mig_type
      call seperr('Not supported mig_type')
    end select

!if((omp_get_thread_num() == 0 )) .and. &
!    i_k_hx == n_hx-3 .and. &
!    i_k_mx == 3 .and. &
!    i_k_my == 3 &
!    ) then
!  write(0,*)'alpha_min=',alpha_min(i_k_mx,i_k_hx)
!  write(0,*)'alpha_max=',alpha_max(i_k_mx,i_k_hx)
!  write(0,*)'k_hy_min=',k_hy_min
!  write(0,*)'k_hy_max=',k_hy_max
!  write(0,*)'d_k_hy=',d_k_hy
!  write(0,*)'k_hy_mean=',k_hy_mean
!
!   end if

   do i_k_hy=1,n_hy
   !{
!      k_hy=(alpha(i_k_mx,i_k_hx)*k_my(i_k_my) + delta_k_hy(i_k_hy))
      k_hy=k_hy_mean + (kk_hy(i_k_hy))*d_k_hy

!if((omp_get_thread_num() == 0) .and. &
!    i_k_hx == n_hx-3 .and. &
!    i_k_mx == 3 .and. &
!    i_k_my == 3 &
!    ) then
!  write(0,*)'i_k_hy=',i_k_hy,'k_hy=',k_hy
!   end if

      ! evaluate DSR with common-azimuth k_hy

!      alg_dist=(i_k_hy-1)
!      if (alg_dist .gt. (n_hy-1)/2) alg_dist=alg_dist-n_hy
!      k_sq_comaz=k_s_x_sq(i_k_mx,i_k_hx) + &
!      ( k_my(i_k_my)*sqrt(alpha(i_k_mx,i_k_hx))+alg_dist*d_k_hy/2. )**2

!      k_sq_comaz=k_s_x_sq(i_k_mx,i_k_hx)+((k_my(i_k_my))**2)*alpha(i_k_mx,i_k_hx)
!      k_sq_comaz=k_s_x_sq(i_k_mx,i_k_hx)+ &
!                 ((k_my(i_k_my))**2)*alpha_s(i_k_mx,i_k_hx)

      k_sq_comaz=k_s_x_sq(i_k_mx,i_k_hx)+ &
                 .25*(k_my(i_k_my) - k_hy)**2

      i_ksq_norm_s=nint(min(k_sq_comaz*norm_ksq+1,max_disp_rel))

!      rphase=phase(i_ksq_norm)*k0_dz
!      rmagn=abs(k0_dz)*magn(i_ksq_norm)

!      k_sq_comaz=k_r_x_sq(i_k_mx,i_k_hx)+ &
!                 ((k_my(i_k_my))**2)*alpha_r(i_k_mx,i_k_hx)

      k_sq_comaz=k_r_x_sq(i_k_mx,i_k_hx)+ &
                 .25*(k_my(i_k_my) + k_hy)**2

      i_ksq_norm_r=nint(min(k_sq_comaz*norm_ksq+1,max_disp_rel))

!      rphase=rphase + phase(i_ksq_norm)*k0_dz
!      rmagn=rmagn + abs(k0_dz)*magn(i_ksq_norm)
      
      ! apply phase shift
!      temp_data_1(i_k_mx,i_k_my,i_k_hx,i_k_hy) =  &
!           temp_data_1(i_k_mx,i_k_my,i_k_hx,i_k_hy)*cexp(cmplx(rmagn,rphase))
!      data_slice(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w) =  &
!           data_slice(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w)*cexp(cmplx(rmagn,rphase))
      data_slice(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w) =  &
           data_slice(i_k_mx,i_k_my,i_k_hx,i_k_hy,i_w)* &
           ph_shift_tab(i_ksq_norm_s)*ph_shift_tab(i_ksq_norm_r)
   !}
   end do
   !}
   end do
   !}
   end do
   !}
   end do

if(apply_any_bound) then
write(0,*)'NAM_mod inside scale 1d=',fft_1d_id%scale,omp_get_thread_num()
  call fft_4d_two_pass_data( &
       fft_1d_id,fft_3d_id,FFT_BACK,n_mx,n_my,n_hx,n_hy,data_slice(:,:,:,:,i_w))

! applyy twice to be cnsinstent with other modules, Once for sources and the other for receivers
  call apply_bound(data_par,data_slice(:,:,:,:,i_w))
  call apply_bound(data_par,data_slice(:,:,:,:,i_w))
write(0,*)'##30'

  call fft_4d_two_pass_data( &
       fft_1d_id,fft_3d_id,FFT_FORW,n_mx,n_my,n_hx,n_hy,data_slice(:,:,:,:,i_w))
end if
   

!} end loop over slowness for sources
end do


!} end loop over frequencies
end do
!''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

!if(prop_data) then
!   deallocate (temp_data_1,stat=ierr)  
!end if

stat=0

return	
end function NAM_down_oper


!-----------------------------------------------------------------------------
function NAM_scatter_oper(adj,add, &
     n_w_data,w_0_data,d_w_data, &
     z_down,d_z_down,data_par, &
     data_slice,ddata_slice,dslow_slice,slow_slice) &
     result(stat)


  integer stat,n_w_data
  logical add,adj
  real w_0_data,d_w_data,z_down,d_z_down
  type (data_param_type) data_par
  real, dimension (:,:,:,:,:) :: slow_slice
  complex, dimension (:,:,:,:,:) :: dslow_slice
  complex, dimension (:,:,:,:,:) :: data_slice,ddata_slice
  optional slow_slice
  
  call seperr('NAM_scatter_oper not supported yet!')
  
  stat=1
  return	
end function NAM_scatter_oper

end module NAM_mod
